# -*- coding: utf-8 -*-
from subprocess import CalledProcessError, run

import pytest

from . import ASSETS_FILES, ASSETS_ENV_FILES


def test_process_excel_file(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "happy_path.xlsx",
            "--output_folder",
            tmpdir,
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "happy_path_sorted.xlsx").exists()


def test_process_excel_file_apply_env_file(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "happy_path.xlsx",
            "--output_folder",
            tmpdir,
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "--env_file",
            ASSETS_ENV_FILES / "default.env",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "happy_path_sorted.xlsx").exists()


def test_process_excel_file_run_twice(tmpdir):
    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "happy_path.xlsx",
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "-o",
            tmpdir,
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "happy_path_sorted.xlsx").exists()

    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "happy_path.xlsx",
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
            "--output_folder",
            tmpdir,
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")
    assert (tmpdir / "happy_path_sorted_1.xlsx").exists()


def test_process_excel_file_using_invalid_definition_file():
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "process-excel-file",
                ASSETS_FILES / "happy_path.xlsx",
                "--excel_definition_file",
                ASSETS_FILES / "excel_definition_invalid_structure.txt",
                "--sprint_schedule_file",
                ASSETS_FILES / "sprint_schedule.json",
            ],
            capture_output=True,
            check=True,
        )

    assert "The structure of excel definition file is wrong" in str(e.value.output)


def test_process_excel_file_using_wrong_input_file():
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "process-excel-file",
                ASSETS_FILES / "excel_definition_invalid_index.json",
                "--excel_definition_file",
                ASSETS_FILES / "excel_definition_invalid_structure.txt",
            ],
            capture_output=True,
            check=True,
        )

    assert "Please provide an Excel file" in str(e.value.output)


def test_process_excel_file_input_file_not_found():
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "process-excel-file",
                ASSETS_FILES / "not_found.xlsx",
                "--excel_definition_file",
                ASSETS_FILES / "excel_definition_invalid_structure.txt",
            ],
            capture_output=True,
            check=True,
        )

    assert "Input file is not exist" in str(e.value.output)


def test_process_excel_file_output_folder_not_exist(tmpdir):
    tmpdir.remove(ignore_errors=True)

    result = run(
        [
            "process-excel-file",
            ASSETS_FILES / "happy_path.xlsx",
            "--output_folder",
            tmpdir,
            "--excel_definition_file",
            ASSETS_FILES / "excel_definition_avoid_jira_operations.json",
            "--sprint_schedule_file",
            ASSETS_FILES / "sprint_schedule.json",
        ],
        capture_output=True,
        check=True,
    )

    assert "xlsx has been saved" in result.stdout.decode("utf-8")


def test_generate_template_excel_definition(tmpdir):
    result = run(
        ["generate-template", "--output_folder", tmpdir, "excel-definition"],
        capture_output=True,
        check=True,
    )

    assert "Generate success" in result.stdout.decode("utf-8")
    assert "excel-definition" in result.stdout.decode("utf-8")


def test_generate_template_excel(tmpdir):
    result = run(
        ["generate-template", "--output_folder", tmpdir, "excel"],
        capture_output=True,
        check=True,
    )

    assert "Generate success" in result.stdout.decode("utf-8")
    assert "excel" in result.stdout.decode("utf-8")


def test_generate_template_sprint_schedule(tmpdir):
    result = run(
        ["generate-template", "sprint-schedule", "--output_folder", tmpdir],
        capture_output=True,
        check=True,
    )

    assert "Generate success" in result.stdout.decode("utf-8")
    assert "schedule" in result.stdout.decode("utf-8")


def test_generate_template_jira_field_mapping_failed(tmpdir):
    with pytest.raises(CalledProcessError) as e:
        run(
            [
                "generate-template",
                "--output_folder",
                tmpdir,
                "jira-field-mapping",
                "--env_file",
                ASSETS_ENV_FILES / "default.env",
            ],
            capture_output=True,
            check=True,
        )

    assert "Generate failed! Template type: jira-field-mapping." in str(e.value.stdout)


def test_generate_template_failed():
    with pytest.raises(CalledProcessError) as e:
        run(["generate-template", "abc"], capture_output=True, check=True)

    assert "argument template_type: invalid choice" in str(e.value.stderr)


def test_update_jira_info():
    result = run(
        [
            "update-jira-info",
            "--access_token",
            "123",
            "--url",
            "http://localhost",
            "--env_file",
            ASSETS_ENV_FILES / "default.env",
        ],
        capture_output=True,
        check=True,
    )

    assert "Add/Update jira url success" in result.stdout.decode("utf-8")
    assert "Add/Update jira access token success" in result.stdout.decode("utf-8")


def test_update_jira_info_failed():
    with pytest.raises(CalledProcessError) as e:
        run(
            ["update-jira-info", "--access_token", " "],
            capture_output=True,
            check=True,
        )

    assert "Please check the access token" in str(e.value.output)



# -*- coding: utf-8 -*-
from __future__ import annotations

from pathlib import Path
from typing import Union


def is_absolute_path_valid(path: Union[str, Path]) -> bool:
    if path is None or not Path(path).is_absolute() or not Path(path).exists():
        return False
    return True



# -*- coding: utf-8 -*-
from jira_assistant.sprint_schedule import SprintScheduleStore

from . import SRC_ASSETS


def test_load_sprint_schedule():
    schedule_filename = SRC_ASSETS / "sprint_schedule.json"
    store = SprintScheduleStore()
    store.load_file(schedule_filename)
    assert store.total_count() > 0


def test_get_priority():
    schedule_filename = SRC_ASSETS / "sprint_schedule.json"
    store = SprintScheduleStore()
    assert store.get_priority("R138") == 0

    store.load_file(schedule_filename)
    assert store.get_priority("R140") == 2



# -*- coding: utf-8 -*-
"""This module will provide console command signatures."""
import pathlib
import warnings

from json import dump
from os import environ, remove
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Tuple

from dotenv import load_dotenv
from urllib3 import disable_warnings

from .excel_definition import ExcelDefinition
from .excel_operation import output_to_excel_file, read_excel_file
from .jira_client import JiraClient, JiraIssueType
from .sprint_schedule import SprintScheduleStore
from .story import (
    Story,
    sort_stories_by_inline_weights,
    sort_stories_by_property_and_order,
    sort_stories_by_raise_ranking,
)

__all__ = ["run_steps_and_sort_excel_file", "generate_jira_field_mapping_file"]

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()

HERE = pathlib.Path(__file__).resolve().parent
ASSETS = HERE / "assets"
DEFAULT_JSON_IDENT = 4
DEFAULT_SPRINT_SCHEDULE_FILE = ASSETS / "sprint_schedule.json"
DEFAULT_EXCEL_DEFINITION_FILE = ASSETS / "excel_definition.json"


def _clear_env_variables():
    if "JIRA_URL" in environ:
        del environ["JIRA_URL"]
    if "JIRA_ACCESS_TOKEN" in environ:
        del environ["JIRA_ACCESS_TOKEN"]


def _get_jira_client(env_file: Optional[Path] = None) -> Optional[JiraClient]:
    if env_file is None:
        if not load_dotenv(ASSETS / ".env"):
            print(
                """The default env file is missing.
                Please use the update-jira-info command to 
                run any command for creating the file."""
            )
            return None
    else:
        _clear_env_variables()
        if not load_dotenv(env_file):
            print("The env file is invalid. Please double check the env file.")
            return None

    jira_url: Optional[str] = environ.get("JIRA_URL", default=None)
    if jira_url is None or jira_url.isspace() or len(jira_url) == 0:
        print(
            """The jira url is invalid.
            Please use the update-jira-info command to add/update url."""
        )
        return None

    jira_acccess_token = environ.get("JIRA_ACCESS_TOKEN", default=None)
    if (
        jira_acccess_token is None
        or jira_acccess_token.isspace()
        or len(jira_acccess_token) == 0
    ):
        print(
            """The jira access token is invalid.
            Please use the update-jira-info command to add/update token."""
        )
        return None

    jira_client = JiraClient(jira_url, jira_acccess_token)

    if not jira_client.health_check():
        print(
            """The jira access token is revoked.
            Please use the update-jira-info command to add/update token."""
        )
        return None

    print(f"Jira info: {jira_url}")

    return jira_client


def _query_jira_information(
    stories: List[Story],
    excel_definition: ExcelDefinition,
    env_file: Optional[Path] = None,
) -> bool:
    jira_client = _get_jira_client(env_file)

    if jira_client is None:
        return False

    jira_fields = []

    for definition_column in excel_definition.get_columns():
        if not definition_column["query_jira_info"] or (
            definition_column["name"] is None
            or definition_column["jira_field_mapping"] is None
        ):
            continue
        jira_fields.append(
            {
                "name": definition_column["name"],
                "jira_name": definition_column["jira_field_mapping"]["name"],
                "jira_path": definition_column["jira_field_mapping"]["path"],
            }
        )

    jira_query_result = jira_client.get_stories_detail(
        [story["storyId"].strip() for story in stories], jira_fields
    )

    for story in stories:
        story_id: str = story["storyId"].lower().strip()
        if story_id in jira_query_result:
            for jira_field in jira_fields:
                story[jira_field["name"]] = jira_query_result[story_id].get(
                    jira_field["jira_name"], None
                )
        else:
            # Story ID has been changed because of convertion.
            tmp_res = jira_client.get_stories_detail([story_id], jira_fields)
            if len(tmp_res) > 0:
                story["storyId"] = list(tmp_res.keys())[0].upper()
                for jira_field in jira_fields:
                    story[jira_field["name"]] = (
                        list(tmp_res.values())[0]
                        .get(jira_field["jira_name"], "")
                        .upper()
                    )
                print(
                    f"""Story id has been changed.
                    Previous: {story_id.upper()}, 
                    Current: {story['storyId'].upper()}"""
                )
            else:
                print(f"Cannot find related information for story: {story_id}")
                story.need_sort = False
                continue

    return True


def _create_jira_stories(
    stories: List[Story],
    excel_definition: ExcelDefinition,
    env_file: Optional[Path] = None,
) -> bool:
    jira_client = _get_jira_client(env_file)

    if jira_client is None:
        return False

    for story in stories:
        input_project_type: Optional[str] = story["projectType"]
        input_story_type: Optional[str] = story["storyType"]
        if input_project_type is None or input_story_type is None:
            print("Please fulfill ProjectType/StoryType field.")
            continue
        project_type = jira_client.get_project_by_project_name(input_project_type)
        if project_type is None:
            print(f"ProjectType: {input_project_type} is not supported.")
            continue
        issue_type: Optional[JiraIssueType] = None
        for item in jira_client.get_issue_types(project_type["name"]):
            if item["name"].lower() == input_story_type.lower():
                issue_type = item
                break
        if issue_type is None:
            print(f"StoryType: {input_story_type} is not supported.")
            continue
        required_fields = jira_client.get_fields_by_project_id_and_issue_id(
            project_type["id"], issue_type["id"]
        )
        all_fields_exist: bool = True
        new_story_fields: Dict[str, Any] = {}
        has_not_allowed_value: bool = False
        for required_field in required_fields:
            required_field_name = required_field["id"].replace(" ", "").lower()
            excel_column = excel_definition.get_column_by_jira_field_mapping_name(
                required_field_name
            )
            if excel_column is None or excel_column["jira_field_mapping"] is None:
                all_fields_exist = False
                break
            excel_column_name = excel_column["name"]
            if not hasattr(story, excel_column_name):
                all_fields_exist = False
                break
            # TODO: Check value is not None?
            allowed_values = required_field.get("allowedValues")
            if allowed_values:
                current_value = story[excel_column_name]
                if (
                    current_value is None
                    or str(current_value).lower() not in allowed_values
                ):
                    print(
                        f"""{excel_column_name} has not allowed value.
                        Value: {current_value}, ProjectType: {project_type['name']}
                        and StoryType: {input_story_type} missing required fields."""
                    )
                    has_not_allowed_value = True
                    break
            new_story_fields[excel_column["jira_field_mapping"]["path"]] = story[
                excel_column_name
            ]
        # Some special fields
        new_story_fields["project.id"] = project_type["id"]
        if not all_fields_exist:
            print(f"StoryType: {input_story_type} missing required fields.")
            continue
        if has_not_allowed_value:
            continue

        new_story = jira_client.create_story(new_story_fields)
        if new_story is not None:
            story["storyId"] = new_story.key
            print(
                f"""New story created.
                  {jira_client.get_jira_browser_link(new_story.key)}"""
            )
            continue
        return False
    return True


def _run_pre_steps(
    stories: List[Story],
    excel_definition: ExcelDefinition,
    env_file: Optional[Path] = None,
):
    # Execute pre-process steps
    pre_process_steps = excel_definition.get_pre_process_steps()

    for pre_process_step in pre_process_steps:
        pre_process_step_name: str = pre_process_step["name"]
        print(f"Executing step: {pre_process_step_name}...")
        if pre_process_step_name.lower() in "RetrieveJiraInformation".lower():
            need_call_jira_api: bool = False
            for excel_definition_column in excel_definition.get_columns():
                if excel_definition_column["jira_field_mapping"] is not None:
                    need_call_jira_api = True
                    break

            if need_call_jira_api:
                stories_need_call_jira: List[Story] = []
                for story in stories:
                    if story.need_sort and story["storyId"] is not None:
                        stories_need_call_jira.append(story)
                if not _query_jira_information(
                    stories_need_call_jira, excel_definition, env_file
                ):
                    print("Retrieve jira information failed.")
                    return
        elif pre_process_step_name.lower() in "FilterOutStoryWithoutId".lower():
            for story in stories:
                if story["storyId"] is None:
                    story.need_sort = False
        elif pre_process_step_name.lower() in "FilterOutStoryBasedOnJiraStatus".lower():
            for story in stories:
                if story["status"] is not None and story[
                    "status"
                ].upper() in pre_process_step["config"].get("JiraStatuses", []):
                    story.need_sort = False
        elif pre_process_step_name.lower() in "CreateJiraStory".lower():
            stories_need_to_create: List[Story] = []
            for story in stories:
                if story["storyId"] is None or story["storyId"].strip() == "":
                    stories_need_to_create.append(story)

            if stories_need_to_create:
                if not _create_jira_stories(
                    stories_need_to_create, excel_definition, env_file
                ):
                    print("Error occurred when creating Jira stories.")
                    return
        print("Executing finish.")


def _run_sort_logics(
    stories: List[Story], excel_definition: ExcelDefinition
) -> Tuple[List[Story], List[Story]]:
    stories_no_need_sort = []
    stories_need_sort = []

    for story in stories:
        if story.need_sort:
            stories_need_sort.append(story)
        else:
            stories_no_need_sort.append(story)

    # Execute sorting logic.
    sort_strategies = excel_definition.get_sort_strategies()

    for sort_strategy in sort_strategies:
        print(f"Executing {sort_strategy['name']} sorting...")
        if sort_strategy["name"] is None:
            continue
        if sort_strategy["name"].lower() in "InlineWeights".lower():
            stories_need_sort = sort_stories_by_inline_weights(stories_need_sort)
        elif sort_strategy["name"].lower() in "SortOrder".lower():
            sort_stories_by_property_and_order(
                stories_need_sort, excel_definition, sort_strategy["config"]
            )
        elif sort_strategy["name"].lower() in "RaiseRanking".lower():
            stories_need_sort = sort_stories_by_raise_ranking(
                stories_need_sort, excel_definition, sort_strategy["config"]
            )
        print("Executing finish.")

    return (stories_need_sort, stories_no_need_sort)


def run_steps_and_sort_excel_file(
    input_file: Union[str, Path],
    output_file: Union[str, Path],
    excel_definition_file: Optional[Union[str, Path]] = None,
    sprint_schedule_file: Optional[Union[str, Path]] = None,
    over_write: bool = True,
    env_file: Optional[Path] = None,
):
    """
    Sort the excel file and output the result

    :parm input_file:
        The excel file need to be sorted. (Absolute path only)

    :parm output_file:
        The sorted excel file location. (Absolute path only)

    :parm sprint_schedule_file:
        The JSON file which contains the priority list to
        calculate the :py:class:`Milestone`

    :parm excel_definition_file:
        The JSON file which contains the input excel file's structure.

    :parm over_write:
        Whether or not the exist output file will be over-write.
    """
    sprint_schedule = SprintScheduleStore()
    if sprint_schedule_file is None:
        print("Using default sprint schedule...")
        sprint_schedule.load(DEFAULT_SPRINT_SCHEDULE_FILE.read_text(encoding="utf-8"))
    else:
        print("Using custom sprint schedule...")
        sprint_schedule.load_file(sprint_schedule_file)

    excel_definition = ExcelDefinition()
    if excel_definition_file is None:
        print("Using default excel definition...")
        excel_definition.load(DEFAULT_EXCEL_DEFINITION_FILE.read_text(encoding="utf-8"))
    else:
        print("Using custom excel definition...")
        excel_definition.load_file(excel_definition_file)

    validation_result = excel_definition.validate()
    if len(validation_result) != 0:
        print(
            """Validating excel definition failed.
            Please check below information to fix first."""
        )
        for item in validation_result:
            print(item)
        return
    print("Validating excel definition success.")

    excel_columns, stories = read_excel_file(
        input_file, excel_definition, sprint_schedule
    )

    if stories is None or len(stories) == 0:
        print("There are no stories inside the excel file.")
        return

    _run_pre_steps(stories, excel_definition, env_file)

    stories_need_sort, stories_no_need_sort = _run_sort_logics(
        stories, excel_definition
    )

    output_to_excel_file(
        output_file,
        stories_need_sort + stories_no_need_sort,  # First output the sorted stories.
        excel_definition=excel_definition,
        columns_in_excel=excel_columns,
        over_write=over_write,
    )

    print(f"{output_file} has been saved.")


def generate_jira_field_mapping_file(
    file: Union[str, Path], over_write: bool = True, env_file: Optional[Path] = None
) -> bool:
    """Generate jira field mapping"""
    jira_client = _get_jira_client(env_file)

    if jira_client is None:
        return False

    output_file_path: Path = Path(file).absolute()

    if output_file_path.exists():
        if over_write:
            try:
                remove(file)
            except PermissionError as e:
                raise FileExistsError(
                    f"The exist jira field mapping file: {file} cannot be removed."
                ) from e
        else:
            raise FileExistsError(
                f"The jira field mapping file: {file} is already exist."
            )

    with open(output_file_path, mode="x", encoding="utf-8") as output_file:
        dump(jira_client.get_all_fields(), output_file, indent=DEFAULT_JSON_IDENT)
        output_file.flush()
        output_file.close()

    return True



# -*- coding: utf-8 -*-
"""
This module is used to store excel column definition information.
"""
import pathlib
import warnings

from json import loads
from sys import version_info

from typing import Any, Dict, List, Optional, TypedDict, Tuple

from jira import JIRA, JIRAError, Issue
from urllib3 import disable_warnings

if version_info < (3, 11):
    from typing_extensions import NotRequired, Self
else:
    from typing import NotRequired, Self


# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()

HERE = pathlib.Path(__file__).resolve().parent
ASSETS = HERE / "assets"
DEFAULT_JIRA_FIELD_TYPE_FILE = ASSETS / "jira_field_type.json"


class JiraFieldTypeDefinition(TypedDict):
    type: NotRequired[str]
    name: NotRequired[str]
    properties: NotRequired[List[Self]]
    isBasic: NotRequired[bool]
    arrayItemType: NotRequired[str]


_jira_field_types = []
_current_jira_field_types_file_path: pathlib.Path = DEFAULT_JIRA_FIELD_TYPE_FILE
# TODO: If file path changed, then reload.


def _init_jira_field_types(jira_field_type_file: Optional[pathlib.Path]):
    if not _jira_field_types or (
        jira_field_type_file is not None
        and not _current_jira_field_types_file_path.samefile(jira_field_type_file)
    ):
        if jira_field_type_file is None:
            jira_field_type_file = DEFAULT_JIRA_FIELD_TYPE_FILE
        for i in loads(jira_field_type_file.read_text(encoding="utf-8")):
            _jira_field_types.append(i)


def get_jira_field(
    field_type: Optional[str], jira_field_type_file: Optional[pathlib.Path] = None
) -> Optional[JiraFieldTypeDefinition]:
    if field_type is None or len(field_type.strip()) == 0:
        return None
    _init_jira_field_types(jira_field_type_file)
    for jira_field_type in _jira_field_types:
        if jira_field_type.get("type", "").lower() == field_type.lower():
            return jira_field_type
    return None


class JiraFieldPropertyPathDefinition(TypedDict):
    path: str
    isArray: bool


def get_field_paths_of_jira_field(
    field_type: str,
    field_property_name: str,
    jira_field_type_file: Optional[pathlib.Path] = None,
) -> Optional[List[JiraFieldPropertyPathDefinition]]:
    jira_field = get_jira_field(field_type, jira_field_type_file)
    if jira_field is None:
        return None
    if jira_field.get("isBasic", False) is True:
        return [{"path": field_property_name, "isArray": False}]
    result = []
    is_array_item = "arrayItemType" in jira_field
    # Following code will use the same jira field type file, so no need to pass.
    _internal_get_field_paths_of_jira_field(
        jira_field,
        is_array_item,
        [
            {
                "path": field_property_name,
                "isArray": is_array_item,
            }
        ],
        result,
    )
    return result


def _internal_get_field_paths_of_jira_field(
    jira_field: Optional[JiraFieldTypeDefinition],
    is_array_item: bool,
    temp: List[JiraFieldPropertyPathDefinition],
    final: List[JiraFieldPropertyPathDefinition],
):
    if jira_field is None:
        return None
    if jira_field.get("isBasic", False) is True:
        for item in temp:
            final.append(
                {
                    "path": connect_jira_field_path(
                        item["path"], jira_field.get("name", "")
                    ),
                    "isArray": is_array_item,
                }
            )
    if "arrayItemType" in jira_field:
        _internal_get_field_paths_of_jira_field(
            get_jira_field(jira_field["arrayItemType"]), True, temp, final
        )
    if "properties" in jira_field:
        field_properties = jira_field.get("properties", [])
        for field_property in field_properties:
            if field_property.get("arrayItemType", None) is not None:
                for item in temp:
                    item["path"] = connect_jira_field_path(
                        item["path"], field_property.get("name", "")
                    )
                _internal_get_field_paths_of_jira_field(
                    get_jira_field(field_property.get("arrayItemType")),
                    True,
                    temp,
                    final,
                )
            if field_property.get("type", None) is None:
                continue
            child_field = get_jira_field(field_property.get("type"))
            if child_field is None:
                continue
            child_field_is_basic = child_field.get("isBasic", False)
            if child_field_is_basic:
                for item in temp:
                    final.append(
                        {
                            "path": connect_jira_field_path(
                                item["path"], field_property.get("name", "")
                            ),
                            "isArray": is_array_item,
                        }
                    )
                continue
            for item in temp:
                item["path"] = connect_jira_field_path(
                    item["path"], field_property.get("name", "")
                )
            _internal_get_field_paths_of_jira_field(
                child_field, is_array_item, temp, final
            )
    return None


def connect_jira_field_path(path_a: str, path_b: str) -> str:
    return path_a + "." + path_b


class JiraProject(TypedDict):
    id: int
    name: str


class JiraIssueType(TypedDict):
    id: int
    name: str
    projectId: int


class JiraField(TypedDict):
    required: bool
    isArray: bool
    name: str
    id: str
    # TODO: Some allowed values are not str.
    allowedValues: NotRequired[List[str]]


# Jira are case-sensitive APIs.
class JiraClient:
    def __init__(self, url: str, access_token: str) -> None:
        self.jira = JIRA(
            server=url,
            token_auth=access_token,
            timeout=20,
            options={"verify": False},
        )
        self._field_cache: Dict[
            str, Dict[str, Optional[List[JiraFieldPropertyPathDefinition]]]
        ] = {}
        self._project_map: Dict[str, JiraProject] = {}
        # The dict key is project_name
        self._project_issue_map_using_name: Dict[str, List[JiraIssueType]] = {}
        # The dict key is project_id
        self._project_issue_map_using_id: Dict[int, List[JiraIssueType]] = {}
        # The dict key is: (project_id, issue_id).
        self._project_issue_field_map: Dict[Tuple[int, int], List[JiraField]] = {}

    def health_check(self) -> bool:
        try:
            if self.jira.myself() is not None:
                return True
            return False
        except JIRAError:
            return False

    # TODO: Based on jira field mapping path to build the dict.
    def create_story(self, fields: Dict[str, Any]) -> "Optional[Issue]":
        try:
            create_issue_body = convert_fields_to_create_issue_body(fields)
            return self.jira.create_issue(
                fields=create_issue_body,
                prefetch=False,
            )
        except JIRAError as e:
            print(
                f"""Calling JIRA API failed.
                HttpStatusCode: {e.status_code}. Response: {e.response.json()}"""
            )
        return None

    def get_jira_browser_link(self, key: str) -> "str":
        return f"{self.jira.server_url}/browse/{key}"

    def get_project_by_project_name(self, project_name: str) -> "Optional[JiraProject]":
        project_name = project_name.strip().lower()
        result = self._project_map.get(project_name, None)
        if result is None:
            self.get_projects()
        return self._project_map.get(project_name, None)

    def get_projects(
        self, include_archived: bool = False, force_refresh: bool = False
    ) -> "List[JiraProject]":
        # Otherwise, if there is no project,
        # still will call API to retrieve project list.
        if self._project_map and force_refresh is not True:
            return list(self._project_map.values())
        self._project_map.clear()
        project_response = self.jira.projects()
        for proj in project_response:
            if include_archived or proj.archived is False:
                proj_name: str = proj.key
                proj_id: int = proj.id
                if proj_name:
                    self._project_map[proj_name.strip().lower()] = {
                        "id": proj.id,
                        "name": proj.key,
                    }
                    # loading project related issue types
                    try:
                        response = self.jira.createmeta_issuetypes(proj_id)
                        total = response.get("total", 0)
                        max_results = response.get("maxResults", 0)
                        if total > max_results:
                            raise NotImplementedError(
                                """Please check 
                                https://github.com/pycontribs/jira/pull/1729 
                                for more info."""
                            )
                        issue_types: List[JiraIssueType] = [
                            {
                                "id": issue_type["id"],
                                "name": issue_type.get("name", ""),
                                "projectId": proj.id,
                            }
                            for issue_type in response.get("values", [])
                        ]
                        self._project_issue_map_using_name[
                            proj_name.strip().lower()
                        ] = issue_types
                        self._project_issue_map_using_id[proj_id] = issue_types
                    except JIRAError:
                        print(f"Get issue types for project: {proj_name} failed.")
                        continue
        return list(self._project_map.values())

    def get_issue_type_by_project_id_and_issue_name(
        self, project_id: int, issue_name: str
    ) -> "Optional[JiraIssueType]":
        match_result = [
            i
            for i in self._project_issue_map_using_id[project_id]
            if i["name"].strip().lower() == issue_name.strip().lower()
        ]
        if match_result:
            return match_result[0]
        return None

    def get_issue_type_by_project_name_and_issue_name(
        self, project_name: str, issue_name: str
    ) -> "Optional[JiraIssueType]":
        project_name = project_name.strip().lower()
        if project_name not in self._project_issue_map_using_name:
            return None
        match_result = [
            i
            for i in self._project_issue_map_using_name[project_name]
            if i["name"].strip().lower() == issue_name.strip().lower()
        ]
        if match_result:
            return match_result[0]
        return None

    def get_issue_types(self, project_name: str) -> "List[JiraIssueType]":
        project_name = project_name.strip().lower()
        result = self._project_issue_map_using_name.get(project_name, [])
        if not result:
            self.get_project_by_project_name(project_name)
        return self._project_issue_map_using_name.get(project_name, [])

    def get_fields_by_project_id_and_issue_id(
        self, project_id: int, issue_id: int, required: bool = True
    ) -> "List[JiraField]":
        result = self._project_issue_field_map.get(
            (project_id, issue_id),
            [],
        )
        if not result:
            issue_response = self.jira.createmeta_issuetypes(project_id)
            total = issue_response.get("total", 0)
            max_results = issue_response.get("maxResults", 0)
            if total > max_results:
                raise NotImplementedError(
                    """Please check 
                    https://github.com/pycontribs/jira/pull/1729 
                    for more info."""
                )
            issue_types = issue_response.get("values", [])
            # Loading all issue types and related fields.
            for issue_type in issue_types:
                # Should be same as issue_name
                issue_type_id: int = issue_type.get("id", None)
                if (
                    project_id,
                    issue_type_id,
                ) not in self._project_issue_field_map:
                    field_response = self.jira.createmeta_fieldtypes(
                        project_id, issue_type_id
                    )
                    total = field_response.get("total", 0)
                    max_results = field_response.get("maxResults", 0)
                    if total > max_results:
                        raise NotImplementedError(
                            """Please check 
                            https://github.com/pycontribs/jira/pull/1729 
                            for more info."""
                        )
                    field_types = field_response.get("values", [])
                    if field_types:
                        self._project_issue_field_map[(project_id, issue_type_id)] = [
                            self._convert_field_type_to_jira_field(field_type)
                            for field_type in field_types
                        ]
        # Try to search again.
        result = self._project_issue_field_map.get(
            (project_id, issue_id),
            [],
        )
        return [i for i in result if i["required"] is required]

    def _convert_field_type_to_jira_field(self, field_type: Any) -> "JiraField":
        result: JiraField
        schema: Dict = field_type.get("schema")
        is_array: bool = "items" in schema
        if is_array:
            allowed_value_type: str = schema.get("items", "")
        else:
            allowed_value_type: str = schema.get("type", "")
        result = {
            "name": field_type.get("name", ""),
            "id": field_type.get("fieldId", ""),
            "isArray": field_type["schema"]["type"] == "array",
            "required": field_type["required"],
        }
        if "allowedValues" in field_type:
            allowed_value_path: str = "value"
            if allowed_value_type == "issuetype":
                allowed_value_path = "name"
            result["allowedValues"] = [
                str(allowed_value.get(allowed_value_path, "")).lower()
                for allowed_value in field_type.get("allowedValues", [])
            ]
        return result

    def get_all_fields(
        self,
    ) -> "Dict[str, Dict[str, Optional[List[JiraFieldPropertyPathDefinition]]]]":
        if not self._field_cache:
            for field in self.jira.fields():
                if "schema" not in field.keys():
                    continue

                temp: Dict[str, Optional[List[JiraFieldPropertyPathDefinition]]] = {
                    "id": field["id"],
                }

                class FieldSchema(TypedDict):
                    type: str
                    items: NotRequired[str]
                    custom: NotRequired[str]
                    customId: NotRequired[int]
                    system: NotRequired[str]

                schema: FieldSchema = field["schema"]
                property_name = field["id"]
                is_array = "items" in schema
                if is_array:
                    field_type = schema.get("items", None)
                else:
                    field_type = schema.get("type", None)

                if field_type is not None:
                    temp["properties"] = get_field_paths_of_jira_field(
                        field_type, property_name
                    )

                    self._field_cache[field["name"]] = temp
        return self._field_cache

    def get_stories_detail(
        self, story_ids: List[str], jira_fields: List[Dict[str, str]]
    ) -> "Dict[str, Dict[str, str]]":
        final_result = {}
        batch_size = 200

        try:
            if len(story_ids) > batch_size:
                start_index = 0
                end_index = batch_size
                while end_index <= len(story_ids) and start_index < len(story_ids):
                    # print(f"Start: {start_index}, End: {end_index}")
                    final_result.update(
                        self._internal_get_stories_detail(
                            story_ids[start_index:end_index], jira_fields
                        )
                    )
                    start_index = end_index
                    if start_index + batch_size < len(story_ids):
                        end_index = start_index + batch_size
                    else:
                        end_index = start_index + (len(story_ids) - end_index)
                return final_result

            return self._internal_get_stories_detail(story_ids, jira_fields)
        except JIRAError as e:
            print(
                f"""Calling JIRA API failed.
                HttpStatusCode: {e.status_code}. Response: {e.response.json()}"""
            )

            return {}

    def _internal_get_stories_detail(
        self, story_ids: List[str], jira_fields: List[Dict[str, str]]
    ) -> "Dict[str, Dict[str, str]]":
        id_query = ",".join([f"'{str(story_id).strip()}'" for story_id in story_ids])

        try:
            search_result: Dict[str, Any] = self.jira.search_issues(
                jql_str=f"id in ({id_query})",
                maxResults=len(story_ids),
                fields=[field["jira_name"] for field in jira_fields],
                json_result=True,
            )  # type: ignore

            final_result = {}
            for issue in search_result["issues"]:
                fields_result = {}
                for field in jira_fields:
                    # First element in the tuple is jira
                    # field name like "customfield_13210 or status..."
                    field_name = field["jira_name"]
                    # Remain elements represent the property path.
                    # Maybe no fields.
                    if "fields" in issue:
                        field_value: Any = issue["fields"]
                        for field_path in field["jira_path"].split("."):
                            if field_value is None:
                                field_value = ""
                                break
                            field_value = field_value.get(field_path, None)
                        fields_result[field_name] = field_value
                final_result[issue["key"].lower()] = fields_result

            return final_result
        except JIRAError as e:
            print(
                f"""Calling JIRA API failed.
                HttpStatusCode: {e.status_code}. Response: {e.response.json()}"""
            )

            return {}


def convert_fields_to_create_issue_body(fields: Dict[str, Any]) -> "Dict[str, Any]":
    issue_fields: Dict[str, Any] = {}
    for key, value in fields.items():
        field_paths = key.split(".")
        tmp = issue_fields
        for count, field_path in enumerate(field_paths):
            if count == len(field_paths) - 1:
                tmp[field_path] = value
            else:
                if tmp.get(field_path, None) is not None:
                    # merge exist dict keys.
                    tmp[field_path] = {**{}, **tmp[field_path]}
                else:
                    tmp[field_path] = {}
            tmp = tmp[field_path]
    return issue_fields



# -*- coding: utf-8 -*-
from __future__ import annotations

from requests_mock import Mocker

from jira_assistant.jira_client import (
    JiraClient,
    get_jira_field,
    get_field_paths_of_jira_field,
    convert_fields_to_create_issue_body,
)
from tests.mock_server import (
    mock_jira_requests,
    mock_jira_requests_with_failed_status_code,
    mock_jira_stories,
)

from . import ASSETS_FILES

DEFAULT_JIRA_URL = "http://localhost"
DEFAULT_JIRA_ACCESS_TOKEN = "123"


def test_get_stories_detail():
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            ["A-1", "A-2", "B-1"],
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 3


def test_get_stories_detail_with_large_amount_of_stories():
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            list(mock_jira_stories.keys()),
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 247


def test_health_check():
    with Mocker(real_http=False, adapter=mock_jira_requests()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        assert client.health_check() is True


def test_health_check_failed():
    with Mocker(real_http=False, adapter=mock_jira_requests_with_failed_status_code()):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        assert client.health_check() is False


def test_get_stories_detail_failed():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests_with_failed_status_code(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            ["A-1", "A-2", "B-1"],
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 0


def test_get_stories_detail_with_large_amount_of_stories_failed():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests_with_failed_status_code(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        stories = client.get_stories_detail(
            list(mock_jira_stories.keys()),
            [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {
                    "name": "status",
                    "jira_name": "status",
                    "jira_path": "status.name",
                },
            ],
        )
        assert len(stories) == 0


def test_get_all_fields():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.get_all_fields()

        assert len(result) == 5


def test_get_projects():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.get_projects()

        assert len(result) == 2
        assert "POC" in [r["name"] for r in result]
        assert "SD" in [r["name"] for r in result]

        result = client.get_projects(include_archived=True, force_refresh=True)
        assert len(result) == 3
        assert "POC" in [r["name"] for r in result]
        assert "SD" in [r["name"] for r in result]
        assert "APPSEC" in [r["name"] for r in result]


def test_get_issue_types():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        result = client.get_issue_types("POC")

        assert len(result) == 7
        assert "Release" in [item["name"] for item in result]


def test_get_fields_by_project_name_and_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        project = client.get_project_by_project_name("POC")
        assert project is not None

        issue_type = client.get_issue_type_by_project_name_and_issue_name(
            project["name"], "Story"
        )
        assert issue_type is not None

        result = client.get_fields_by_project_id_and_issue_id(
            project["id"], issue_type["id"]
        )

        assert len(result) == 2


def test_get_fields_by_unkown_project_name_and_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        issue_type = client.get_issue_type_by_project_name_and_issue_name(
            "Proj1", "Story1"
        )
        assert issue_type is None


def test_get_fields_by_project_name_and_unkown_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        project = client.get_project_by_project_name("POC")
        assert project is not None

        issue_type = client.get_issue_type_by_project_name_and_issue_name(
            project["name"], "Story1"
        )
        assert issue_type is None


def test_get_fields_by_project_id_and_issue_name():
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests(),
    ):
        client = JiraClient(DEFAULT_JIRA_URL, DEFAULT_JIRA_ACCESS_TOKEN)

        project = client.get_project_by_project_name("POC")
        assert project is not None

        issue_type = client.get_issue_type_by_project_id_and_issue_name(
            project["id"], "Story"
        )
        assert issue_type is not None

        result = client.get_fields_by_project_id_and_issue_id(
            project["id"], issue_type["id"]
        )

        assert len(result) == 2


def test_convert_fields_to_create_issue_body():
    issue_fields = {
        "issuetype.name": "Release Support",
        "issuetype.id": "1000",
        "summary": "test",
        "project.id": 10100,
        "project.key": "Sandbox",
    }

    create_issue_fields = convert_fields_to_create_issue_body(issue_fields)
    assert create_issue_fields["issuetype"]["name"] == "Release Support"
    assert create_issue_fields["issuetype"]["id"] == "1000"
    assert create_issue_fields["summary"] == "test"
    assert create_issue_fields["project"]["id"] == 10100
    assert create_issue_fields["project"]["key"] == "Sandbox"


def test_get_jira_field_use_basic_type():
    assert get_jira_field(None) is None
    assert get_jira_field("abc") is None
    actual_field = get_jira_field("any")
    assert actual_field is not None
    assert "isBasic" in actual_field
    assert actual_field.get("isBasic", False) is True


def test_get_jira_field_use_complex_type():
    actual_field = get_jira_field("status")
    assert actual_field is not None
    assert "isBasic" in actual_field
    assert actual_field.get("isBasic", True) is False
    assert actual_field.get("type") == "status"
    properties = actual_field.get("properties")
    assert properties is not None
    assert "name" in [p.get("name", "") for p in properties]
    assert "statusCategory" in [p.get("name", "") for p in properties]


def test_get_field_paths_of_jira_field_use_unkown_type():
    actual_field_paths = get_field_paths_of_jira_field(
        "abc", "abc", ASSETS_FILES / "jira_field_type.json"
    )
    assert actual_field_paths is None


def test_get_field_paths_of_jira_field_use_child_array_type():
    actual_field_paths = get_field_paths_of_jira_field(
        "array-level1", "abc", ASSETS_FILES / "jira_field_type.json"
    )
    assert actual_field_paths is not None
    assert "abc.name.level3" in [item["path"] for item in actual_field_paths]


def test_get_field_paths_of_jira_field_use_basic_type():
    actual_field_paths = get_field_paths_of_jira_field("string", "customfield_17001")
    assert actual_field_paths is not None
    assert "customfield_17001" == actual_field_paths[0]["path"]
    assert not actual_field_paths[0]["isArray"]


def test_get_field_paths_of_jira_field_use_complex_type_no_hierarchy():
    # Author
    actual_field_paths = get_field_paths_of_jira_field("author", "abc")
    assert actual_field_paths is not None
    assert len(actual_field_paths) == 2
    assert "abc.name" in [item["path"] for item in actual_field_paths]
    assert "abc.emailAddress" in [item["path"] for item in actual_field_paths]


def test_get_field_paths_of_jira_field_use_complex_type_multiple_hierarchy():
    # Project
    actual_field_paths = get_field_paths_of_jira_field("project", "abc")
    assert actual_field_paths is not None
    assert len(actual_field_paths) == 5
    assert "abc.name" in [item["path"] for item in actual_field_paths]
    assert "abc.key" in [item["path"] for item in actual_field_paths]
    assert "abc.projectTypeKey" in [item["path"] for item in actual_field_paths]
    assert "abc.projectCategory.description" in [
        item["path"] for item in actual_field_paths
    ]
    assert "abc.projectCategory.name" in [item["path"] for item in actual_field_paths]


def test_get_field_paths_of_jira_field_use_array_type_no_hierarchy():
    # comments-page
    actual_field_paths = get_field_paths_of_jira_field("comments-page", "abc")
    assert actual_field_paths is not None
    assert len(actual_field_paths) == 5
    assert "abc.comments.author.name" in [item["path"] for item in actual_field_paths]
    assert "abc.comments.author.emailAddress" in [
        item["path"] for item in actual_field_paths
    ]
    assert "abc.comments.author.id" in [item["path"] for item in actual_field_paths]
    assert "abc.comments.author.body" in [item["path"] for item in actual_field_paths]
    assert "abc.comments.author.created" in [
        item["path"] for item in actual_field_paths
    ]



# -*- coding: utf-8 -*-
# pylint: disable=line-too-long
from pytest import raises

from jira_assistant.excel_definition import ExcelDefinition

from . import ASSETS_FILES, SRC_ASSETS


def test_load_happy_path():
    excel_definition_filename = SRC_ASSETS / "excel_definition.json"

    store = ExcelDefinition()
    with open(excel_definition_filename, encoding="utf-8") as file:
        store.load(file.read())
    assert store.total_count() > 0


def test_load_using_none_input():
    store = ExcelDefinition()
    content: str = None  # type: ignore
    with raises(ValueError) as err:
        store.load(content)
    assert "There is no content in the excel definition file." in str(err.value)


def test_load_file():
    excel_definition_filename = SRC_ASSETS / "excel_definition.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)
    assert store.total_count() > 0


def test_load_file_none():
    store = ExcelDefinition()
    file: str = None  # type: ignore
    with raises(FileNotFoundError) as err:
        store.load_file(file)
    assert (
        "Please make sure the excel definition file exist and the path should be absolute"
        in str(err.value)
    )


def test_iter():
    excel_definition_filename = SRC_ASSETS / "excel_definition.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)
    items = []
    for item in store:
        items.append(item)
    assert len(items) > 0


def test_validate():
    excel_definition_filename = SRC_ASSETS / "excel_definition.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 0


def test_validate_missing_story_id():
    excel_definition_filename = ASSETS_FILES / "excel_definition_missing_story_id.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_duplicate_index():
    excel_definition_filename = ASSETS_FILES / "excel_definition_duplicate_index.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_duplicate_inline_weights():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_duplicate_inline_weights.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_invalid_name():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_name.json"
    store = ExcelDefinition()
    with raises(SyntaxError) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The excel definition file has below issues need to be fixed"
        in err.value.args[0]
    )
    assert (
        "The Name property type in the column definition should be string."
        in err.value.args[0]
    )


def test_validate_invalid_raise_ranking():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_raise_ranking.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_invalid_require_sort():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_require_sort.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 2


def test_validate_invalid_type():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_type.json"
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 2


def test_validate_invalid_index():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_index.json"
    store = ExcelDefinition()
    with raises(SyntaxError) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The excel definition file has below issues need to be fixed"
        in err.value.args[0]
    )
    assert (
        "The Index property type in the column definition is not integer."
        in err.value.args[0]
    )


def test_validate_index_not_continuation():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_index_not_continuation.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_strategy_duplicate_priority():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_strategy_duplicate_priority.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_strategy_invalid_name():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_strategy_invalid_name.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_invalid_structure():
    excel_definition_filename = ASSETS_FILES / "excel_definition_invalid_structure.txt"
    store = ExcelDefinition()

    with raises(Exception) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The structure of excel definition file is wrong. Hint: Expecting ',' delimiter in line 49:1"
        in err.value.args[0]
    )


def test_validate_invalid_pre_process_step_name():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_pre_process_step_name.json"
    )
    store = ExcelDefinition()

    with raises(SyntaxError) as err:
        store.load_file(excel_definition_filename)

    assert (
        "The excel definition file has below issues need to be fixed"
        in err.value.args[0]
    )
    assert "The pre-process step must have a name." in err.value.args[0]


def test_validate_invalid_pre_process_step_priority():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_invalid_pre_process_step_priority.json"
    )
    store = ExcelDefinition()

    store.load_file(excel_definition_filename)
    validation_result = store.validate()

    assert len(validation_result) == 1


def test_validate_strategy_duplicate_column_story_id():
    excel_definition_filename = (
        ASSETS_FILES / "excel_definition_duplicate_column_story_id.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert len(validation_result) == 1
    assert (
        "Column named StoryId has been duplicated. Column: storyId"
        in validation_result[0]
    )


def test_validate_invalid_pre_process_step_create_jira_story():
    excel_definition_filename = (
        ASSETS_FILES
        / "excel_definition_invalid_pre_process_step_create_jira_story.json"
    )
    store = ExcelDefinition()
    store.load_file(excel_definition_filename)

    validation_result = store.validate()

    assert (
        "The PreProcessStep: CreateJiraStory must have a column named ProjectType."
        in validation_result[0]
    )



# -*- coding: utf-8 -*-
import pathlib
from os import environ
import pytest

from requests_mock import Mocker

from jira_assistant.assistant import (
    generate_jira_field_mapping_file,
    run_steps_and_sort_excel_file,
)
from jira_assistant.excel_definition import ExcelDefinition
from jira_assistant.excel_operation import read_excel_file
from jira_assistant.jira_client import JiraClient
from jira_assistant.sprint_schedule import SprintScheduleStore
from jira_assistant.story import compare_story_based_on_inline_weights
from tests.mock_server import (
    mock_jira_requests,
    mock_jira_requests_with_failed_status_code,
)
from tests.utils import read_stories_from_excel

from . import ASSETS_ENV_FILES, ASSETS_FILES, SRC_ASSETS


def test_run_steps_and_sort_excel_file(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )

        excel_definition = ExcelDefinition()
        excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
        sprint_schedule = SprintScheduleStore()
        sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

        _, stories = read_excel_file(
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition,
            sprint_schedule,
        )

        assert len(stories) == 8

        jira_client = JiraClient(environ["JIRA_URL"], environ["JIRA_ACCESS_TOKEN"])

        noneed_sort_statuses = [
            "SPRINT COMPLETE",
            "PENDING RELEASE",
            "PRODUCTION TESTING",
            "CLOSED",
        ]

        jira_fields = [
            {
                "name": "domain",
                "jira_name": "customfield_15601",
                "jira_path": "customfield_15601.value",
            },
            {"name": "status", "jira_name": "status", "jira_path": "status.name"},
        ]

        for i in range(len(stories) - 1):
            story_id_0 = stories[i]["storyId"].lower().strip()
            story_id_1 = stories[i + 1]["storyId"].lower().strip()
            query_result = jira_client.get_stories_detail(
                [story_id_0, story_id_1], jira_fields
            )
            if (
                query_result[story_id_0]["status"].upper() not in noneed_sort_statuses
                and query_result[story_id_1]["status"].upper()
                not in noneed_sort_statuses
            ):
                assert (
                    compare_story_based_on_inline_weights(stories[i], stories[i + 1])
                    >= 0
                )


def test_run_steps_and_sort_excel_file_use_default_files(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            env_file=ASSETS_ENV_FILES / "default.env",
        )

        excel_definition = ExcelDefinition()
        excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
        sprint_schedule = SprintScheduleStore()
        sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

        _, stories = read_excel_file(
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition,
            sprint_schedule,
        )

        assert len(stories) == 8


def test_run_steps_and_sort_excel_file_use_wrong_excel_definition_file(capsys, tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition_file=ASSETS_FILES
            / "excel_definition_duplicate_index.json",
            env_file=ASSETS_ENV_FILES / "default.env",
        )
        output = capsys.readouterr()
        assert "Validating excel definition failed." in output.out


def test_run_steps_and_sort_excel_file_with_empty_excel_file(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "empty_excel.xlsx",
            tmpdir / "empty_excel_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
        )


def test_run_steps_and_sort_excel_file_with_raise_ranking_file(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition_file=str(
                ASSETS_FILES / "excel_definition_with_raise_ranking.json"
            ),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )

        stories = read_stories_from_excel(
            tmpdir / "happy_path_sorted.xlsx",
            SRC_ASSETS / "excel_definition.json",
            SRC_ASSETS / "sprint_schedule.json",
        )

        false_value_begin = False
        for story in stories:
            if story["isThisAHardDate"] is True:
                continue
            if story["isThisAHardDate"] is False and false_value_begin is False:
                false_value_begin = True
                continue
            if story["isThisAHardDate"] is True and false_value_begin is True:
                raise AssertionError


def test_run_steps_and_sort_excel_file_missing_jira_url(capsys, tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "missing_jira_url.env",
        )
        output = capsys.readouterr()
        assert "The jira url is invalid." in output.out


def test_run_steps_and_sort_excel_file_missing_access_token(capsys, tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "missing_jira_access_token.env",
        )
        output = capsys.readouterr()
        assert "The jira access token is invalid." in output.out


def test_run_steps_and_sort_excel_file_jira_health_check_failed(capsys, tmpdir):
    with Mocker(
        real_http=False,
        case_sensitive=False,
        adapter=mock_jira_requests_with_failed_status_code(),
    ):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path.xlsx",
            tmpdir / "happy_path_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )
        output = capsys.readouterr()
        assert "The jira access token is revoked." in output.out


def test_generate_jira_field_mapping_file(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        output_file: pathlib.Path = tmpdir / "jira_field_mapping.json"

        assert (
            generate_jira_field_mapping_file(
                output_file, env_file=ASSETS_ENV_FILES / "default.env"
            )
            is True
        )
        assert output_file.exists() is True


def test_generate_jira_field_mapping_file_over_write_is_true(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        output_file: pathlib.Path = pathlib.Path(tmpdir) / "jira_field_mapping.json"
        output_file.touch(exist_ok=True)

        assert (
            generate_jira_field_mapping_file(
                output_file, over_write=True, env_file=ASSETS_ENV_FILES / "default.env"
            )
            is True
        )
        assert output_file.exists() is True


def test_generate_jira_field_mapping_file_over_write_is_false(tmpdir):
    with pytest.raises(FileExistsError) as e:
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            output_file: pathlib.Path = pathlib.Path(tmpdir) / "jira_field_mapping.json"
            output_file.touch(exist_ok=True)

            generate_jira_field_mapping_file(
                output_file,
                over_write=False,
                env_file=ASSETS_ENV_FILES / "default.env",
            )
    assert "already exist" in str(e.value.args[0])


def test_run_steps_and_sort_excel_file_with_no_need_sort_stories(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path_with_no_need_sort_stories.xlsx",
            tmpdir / "happy_path_with_no_need_sort_stories_sorted.xlsx",
            excel_definition_file=str(ASSETS_FILES / "excel_definition.json"),
            sprint_schedule_file=str(ASSETS_FILES / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )

        excel_definition = ExcelDefinition()
        excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
        sprint_schedule = SprintScheduleStore()
        sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

        _, stories = read_excel_file(
            tmpdir / "happy_path_with_no_need_sort_stories_sorted.xlsx",
            excel_definition,
            sprint_schedule,
        )

        assert len(stories) == 8

        jira_client = JiraClient(environ["JIRA_URL"], environ["JIRA_ACCESS_TOKEN"])

        noneed_sort_statuses = [
            "SPRINT COMPLETE",
            "PENDING RELEASE",
            "PRODUCTION TESTING",
            "CLOSED",
        ]

        jira_fields = [
            {
                "name": "domain",
                "jira_name": "customfield_15601",
                "jira_path": "customfield_15601.value",
            },
            {"name": "status", "jira_name": "status", "jira_path": "status.name"},
        ]

        for i in range(len(stories) - 1):
            if stories[i]["storyId"] is None or stories[i + 1]["storyId"] is None:
                continue
            story_id_0 = stories[i]["storyId"].lower().strip()
            story_id_1 = stories[i + 1]["storyId"].lower().strip()
            query_result = jira_client.get_stories_detail(
                [story_id_0, story_id_1], jira_fields
            )
            if (
                query_result[story_id_0]["status"].upper() not in noneed_sort_statuses
                and query_result[story_id_1]["status"].upper()
                not in noneed_sort_statuses
            ):
                assert (
                    compare_story_based_on_inline_weights(stories[i], stories[i + 1])
                    >= 0
                )


def test_create_jira_stories_and_sort_excel_file(tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path_create_story.xlsx",
            tmpdir / "happy_path_create_story_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )

        excel_definition = ExcelDefinition()
        excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
        sprint_schedule = SprintScheduleStore()
        sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

        _, stories = read_excel_file(
            tmpdir / "happy_path_create_story_sorted.xlsx",
            excel_definition,
            sprint_schedule,
        )

        assert len(stories) == 1


def test_create_jira_stories_invalid_project_type(capsys, tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path_create_story_invalid_project_type.xlsx",
            tmpdir / "happy_path_create_story_invalid_project_type_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )
        output = capsys.readouterr()
        assert "ProjectType: ABC is not supported." in output.out


def test_create_jira_stories_invalid_issue_type(capsys, tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path_create_story_invalid_issue_type.xlsx",
            tmpdir / "happy_path_create_story_invalid_issue_type_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )
        output = capsys.readouterr()
        assert "StoryType: WHH is not supported." in output.out


# def test_create_jira_stories_missing_required_field(capsys, tmpdir):
#    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
#        run_steps_and_sort_excel_file(
#            ASSETS_FILES / "happy_path_create_story_missing_required_field.xlsx",
#            tmpdir / "happy_path_create_story_missing_required_field_sorted.xlsx",
#            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
#            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
#            env_file=ASSETS_ENV_FILES / "default.env",
#        )
#        output = capsys.readouterr()
#        assert "StoryType: sd missing required fields." in output.out


def test_create_jira_stories_no_project_type(capsys, tmpdir):
    with Mocker(real_http=False, case_sensitive=False, adapter=mock_jira_requests()):
        run_steps_and_sort_excel_file(
            ASSETS_FILES / "happy_path_create_story_no_project_type.xlsx",
            tmpdir / "happy_path_create_story_no_project_type_sorted.xlsx",
            excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
            sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            env_file=ASSETS_ENV_FILES / "default.env",
        )
        output = capsys.readouterr()
        assert "Please fulfill ProjectType/StoryType field." in output.out
